// Here we store the data for this App
var DATA_CONFIG = 'imatch.notes.{u}.config';

// The name of the Attribute set and the Attribute we use when 
// the user wants to store notes in Attributes.
// We create this set when it does not exist.
var NOTES_ATTR_SET = 'IMatchNotes';
var NOTES_ATTR = 'Note';

// This object holds the configuration data for this.
var config = {
    version : 1,
    storageType: 'st-xmp-desc',
    fontSize: 'fs-2',
    fontStyle: 1,       // 1 = fancy font
}

// This object holds the state and data at the time we loaded the note.
// We may save the note _after_ the user has changed the focused file or selection
// and in that case we need to know the focused file and selection at the time the note 
// was loaded.
var loadState = {
    focusedFile : 0,                    // Focused file at load time
    instanceId : 0,                     // Attribute: The instance id for which we show the data
    selectedFiles: [],                  // The ids of the selected files
    data : ''                           // The data we loaded (to know if something was changed)
}

var writeBackData = null;               // The data to write (set whenever the user changes the <input>)

//////////////////////////////////////////////////////////////////
//
$(document).ready(function () {

    // Translate the app
    IMatchTranslate.init(IMatchTranslate.getUserLanguage(),'en','res/');
    //IMatchTranslate.init('de','en','res/');

    // Request the user info first. We need the user name to load all other settings.
    IMatch.userInfo().then(function(response) {

        DATA_CONFIG = DATA_CONFIG.replace('{u}',response.userName);

        // Now load the settings
        IMWS.dataGet({name: DATA_CONFIG}).then(function(response) {
            var data = JSON.parse(response.value);
            if (data.version == 1) {
                config = data;
            }
            applyConfig();

            // Make sure the Attribute set exists when we use it
            checkNotesSet().always(function() {
                loadNote();
            });
        },function() {
            // No configuraton found, show the config pverlay
            applyConfig();
            toggleConfig(true);
        });
    });

    // Button handlers

    $('#btn-config').click(function() {
        toggleConfig(true);
    });

    $('#btn-config-ok').click(function(e) {
        saveConfig();
        toggleConfig(false);
        applyConfig();
        checkNotesSet().always(function() {
            loadNote();
        });
        e.preventDefault();
        return false;
    });

    $('#btn-config-cancel').click(function(e) {
        toggleConfig(false);

        e.preventDefault();
        return false;
    });

    // THe font-size buttons are all handled by this
    $('.fs-class').click(function(e) {
        config.fontSize = $(this).attr('id');
    });

    // This works together with the input handler for the note.
    // We save ns after the user has stopped typing.
    var delayStoreNote = IMatch.debounce(function() {
        storeNote();
    }, 1000);

    // This handles all sorts of input into the Note <textarea>, 
    // including typing, copy/paste etc.
    $('#note').on('input',function(e) {
        writeBackData = $('#note').val();
        delayStoreNote();
    });

    // Setup event monitoring
    setupMonitoring();       

}); // document.ready


///////////////////////////////////////////////////////////////////
// We monitor IMatch and IMWS events.
function setupMonitoring()
{
    var delayLoadNote = IMatch.debounce(function() {
        // If we have a pending write-back, first store, then load.
        if (writeBackData != null) {
            storeNote().then(function() {
                loadNote();
            });
        }
        else {
            loadNote();
        }
    }, 500);
    
    if (IMatch.isEmbedded()) {
        IMatch.openWebSocket();
    }
    else {
        // Running in the browser
        IMatch.openWebSocket(IMatch.DEFAULT_WS_URL);
    }

    if (IMatch.isEmbedded()) {
        IMWS.openWebSocket();
    }
    else {
        // Running in the browser
        IMWS.openWebSocket(IMatch.DEFAULT_WS_URL);
    }

    // Monitor IMatch events to get notice about file window events.
    IMatch.webSocket().onmessage = function(e) {
        var msg = JSON.parse(e.data);
        if (msg.msg == 'fileWindow.selectionChanged') {
            delayLoadNote();
        }
        else if (msg.msg == 'fileWindow.focusChanged') {
            delayLoadNote();
        }
    }

    // Monitor IMWS events to get notice when file metadata or attributes change.
    IMWS.webSocket().onmessage = function(e) {
        var msg = JSON.parse(e.data);

        if (msg.ma == IMWS.EVENT_CODES.evObjectUpdate) {
            switch (msg.mi) {
            case IMWS.EVENT_CODES.evOUMBFileValueUpdate:
                if (config.storageType == 'st-xmp-desc') {
                    // The metadata of a file has changed.
                    // If this is the file we are currently showing, we need to reload
                    if (msg.d1 == loadState.focusedFile) {
                        delayLoadNote();
                    }
                }
                break;
            }
        }
        else if (msg.ma == IMWS.EVENT_CODES.evAttributeDB) {
            if (config.storageType == 'st-attr') {
                switch (msg.mi) {
                case IMWS.EVENT_CODES.evSDBDataAdded:
                case IMWS.EVENT_CODES.evSDBDataUpdated:
                case IMWS.EVENT_CODES.evSDBDataDeleted:
                {
                    // Attributes data has changed.
                    // If this is the file we are currently showing, we need to reload
                    if (msg.d1 == loadState.focusedFile) {
                        delayLoadNote();
                    }
                }
                break;
                }
            }
        }
    }
}


//////////////////////////////////////////////////////////////////
// Transfer settings from config to the config form and the note
function applyConfig()
{
    $('#imatch-attribute-set-name').text(NOTES_ATTR_SET);

    $('#' + config.storageType).attr('checked',true);
    $('#' + config.fontSize).button('toggle');

    // Update the font settings for the note textarea

    var n = $('#note');
    // Remove all fs- classes (font size)
    n.removeClass(function(index,className) {
        return className.match(/fs\-1|fs\-2|fs\-3|fs\-4/g) != null ? className : '';
    });
    // Add the class matching the selected font size
    n.addClass(config.fontSize);

    $('#font-handwriting').prop('checked',config.fontStyle == 1 ? true : false);

    // Do we use a non-standard font?
    n.removeClass('handwriting-font');
    if (config.fontStyle == 1) {
        n.addClass('handwriting-font');
    }
}


//////////////////////////////////////////////////////////////////
// Store the current configuraton
function saveConfig()
{
    config.storageType =  $('input:radio[name=storage-type]:checked').attr('id');

    if ($('#font-handwriting').prop('checked')) {
        config.fontStyle = 1;
    }
    else {
        config.fontStyle = 0;
    }

    // Now load the settings
    IMWS.dataSet({
        name: DATA_CONFIG,
        value: JSON.stringify(config)
    });
}


//////////////////////////////////////////////////////////////////
// show/hide the configuration
function toggleConfig(show)
{
    if (show) {
        $('#app-container').fadeOut(200,function() {
            $('#config-container').fadeIn();
        });
    }
    else {
        $('#config-container').fadeOut(200,function() {
            $('#app-container').fadeIn();
        });
    }
}


//////////////////////////////////////////////////////////////////
// Make sure the Attribute set we use exsits
function checkNotesSet()
{
    const IMATCH_ATTRIBUTES = [
        {
            name: NOTES_ATTR,
            type: 'string'
        }
    ]

    var dfd = jQuery.Deferred();

    if (config.storageType == 'st-attr') {
        IMWS.get('v1/attributes/schema',{
            set: NOTES_ATTR_SET
        }).then(function(response) {
            if (response.schema.length == 0) {
                // Create the Set
                IMWS.post('v1/attributes/schema/set/create',{
                    name: NOTES_ATTR_SET,
                    description: 'This Attribute IMatch Notes. DO NOT MANUALLY EDIT!',
                    type: 'file',
                    // We create a hidden set
                    hidden: true,
                    attributes: JSON.stringify(IMATCH_ATTRIBUTES)
                }).then(function(response) {
                    dfd.resolve();
                },function(error) {
                    alert(IMatchTranslate.getStr('T_ATTR_ERROR'));
                    dfd.reject();
                });
            }
            else {
                dfd.resolve();
            }
        },function(error) {
            alert(IMatchTranslate.getStr('T_SCHEMA_ERROR'));
            dfd.reject();
        });
    }
    else {
        dfd.resolve();
    }

    return dfd.promise();
}


//////////////////////////////////////////////////////////////////
// Stores the note in the Attribute or tag.
// Returns a promise so we can wait on this function to complete.
function storeNote()
{
    var dfd = jQuery.Deferred();

    // Nothing to do
    if ((loadState.focusedFile == 0) || (writeBackData == null)) {
        dfd.resolve();
        return dfd;
    }

    $('#spinner').show();

    // We store the data into all files which were selected
    // at the time we loaded the note. 
    // Make sure the focused file is considered even when it is
    // not selected.

    if (loadState.selectedFiles.indexOf(loadState.focusedFile) < 0) {
        loadState.selectedFiles.push(loadState.focusedFile);
    }

    // This is the data we have loaded
    var originalData = loadState.data;

    // Remember the data we have written
    loadState.data = writeBackData;
    writeBackData = null;
        
    if (config.storageType == 'st-xmp-desc') {
                // Store in XMP descrription, replacing whatever is there.
                    
        var tasks = [
            {
                id: loadState.selectedFiles.join(','),
                //tag: 'description',
                //op: 'set',
                tag: 'hierarchicalkeywords',
                op: 'add',
                value: loadState.data
            }
        ]
        IMWS.post('v1/metadata',{
            tasks: JSON.stringify(tasks)
        }).then(function(response) {
        },function(error) {
            alert(IMatchTranslate.getStr('T_STORE_ERROR'));
        }).always(function() {
            $('#spinner').fadeOut();
            dfd.resolve();
        });
    }
    else {
        // NOTE:
        // This is more complex than metadata becase we need to figure out whether to 
        // a) add a new row
        // b) update an existing row
        // If a file already has a record in the note set, we update it, else we add it.

        // In the case that loadState.data is empty (the user has erased the note) we delete
        // the entire existing attribute record.

        var add = [];           // The files for which to add new data
        var update = [];        // The files and instance ids to update

        // We start by loading the Attribute data for all selected files (including the focused)
        IMWS.get('v1/attributes',{
            id: loadState.selectedFiles.join(','),
            set: NOTES_ATTR_SET,
            // We also need records for files without data!
            returnempty: true,
            attributes: [NOTES_ATTR]
        }).then(function(response) {


            // For each file we check if the file has data.
            // If not => add the id to the add array.
            // Else add it to the update array.
            response.result.forEach(function(r) {

                // If the file has no data, we put it into add
                if (r.data.length == 0) {
                    // Only when we have data
                    if (loadState.data.length > 0) {
                        add.push(r.id);
                    }
                }
                else {
                    if (loadState.data.length > 0) {
                        // We remember the file id and the instance id to update
                        update.push({id: r.id, instanceId: r.data[0].instanceId});
                    }
                    else {
                        // For delete we only need the id
                        update.push(r.id);
                    }
                }
            }, this);

            // We use this array to save the promises returned by IMWS.
            // We may need to wait for one or two promises!
            var requests = [];
            
            // Now we know for which files we need to add Attributes and
            // for which files we need to update existing instances.
            if (add.length > 0) {
                var tasks = [{
                    op: 'add',
                    data: {}
                }]
                tasks[0].data[NOTES_ATTR] = loadState.data;

                requests.push(IMWS.post('v1/attributes', {
                    id: add.join(','),
                    set: NOTES_ATTR_SET,
                    tasks: JSON.stringify(tasks)
                }));
            }

            if (update.length > 0) {

                if (loadState.data.length == 0) {
                    var task = {
                        op: 'delete',
                        instanceid: []
                    };
                    var tasks = [];
                    tasks.push(task);

                    requests.push(IMWS.post('v1/attributes', {
                        set: NOTES_ATTR_SET,
                        id: update.join(','),
                        tasks: JSON.stringify(tasks)
                    }));
                }
                else {
                    // We create one task for each file to update, using
                    // the option to specify the file id as part of the task

                    var tasks = [];

                    update.forEach(function(u) {
                        var task = {
                            op: 'update',
                            id: [u.id],
                            instanceid: [u.instanceId],
                            data: {}
                        };
                        task.data[NOTES_ATTR] = loadState.data;
                        tasks.push(task);
                    }, this);

                    requests.push(IMWS.post('v1/attributes', {
                        set: NOTES_ATTR_SET,
                        tasks: JSON.stringify(tasks)
                    }));
                }
            }

            // Wait for all pending promises
            // Note: Then apply syntax allows us to supply an array of deferreds to when.
            //       See: https://stackoverflow.com/questions/5627284/pass-in-an-array-of-deferreds-to-when
            $.when.apply($,requests).always(function() {
                $('#spinner').fadeOut();
                dfd.resolve();
            });
        });
    }

    return dfd.promise();
}


//////////////////////////////////////////////////////////////////
// Stores the note in the Attribute or tag
function loadNote()
{
    loadState.focusedFile = 0;
    loadState.instanceId = 0;
    loadState.selectedFiles = [];
    // Do NOT reset the currently loaded data (if any) yet!

    // First get the id of the focused file from the idlist. This is
    // our anchor and we show its data (if any) to the user.
    IMWS.get('v1/files',{
        idlist: IMatch.idlist.fileWindowFocusedFile,
        fields: 'id,name'
    }).then(function(response) {
        // Do we have a focused file?
        if (response.files.length) {

            // Load and remember the ids of all selected files.
            // Note: The focused file may not be selected!
            IMWS.get('v1/idlist',{
                name: IMatch.idlist.fileWindowSelection
            }).then(function(response) {
                loadState.selectedFiles = response.id;
                // Show how many files are selected if more than 1
                if (loadState.selectedFiles.length > 1) {
                    $('#sel-count').text( '(' + loadState.selectedFiles.length + IMatchTranslate.getStr('T_FILES_SELECTED') + ')');
                }
                else {
                    $('#sel-count').text('');
                }
            });
            
            // Remember the focused file for which we loaded the note data
            // and show the file name to the user.
            loadState.focusedFile = response.files[0].id;
            $('#focused-file').html('<span>' + IMatchTranslate.getStr('T_SHOWING') + ' <strong>' + response.files[0].name + '<strong></span>');

            // Load the metadata or Attribute data for the focused file.
            // This is what we show to the user

            if (config.storageType == 'st-xmp-desc') {
                IMWS.get('v1/files',{
                    id: loadState.focusedFile,
                    fields: 'id',
                    //tagdesc: 'description'
                    //tagdesc: 'hierarchicalkeywords'
                }).then(function(response) {
                    if (response.files.length) {
                        // If the data loaded differs from the last loaded data,
                        // we update the note. This avoids unwanted updates and focus/cursor changes
                        // after storing, when IMWS sends 'changed' events and we react.
                        if (loadState.data != response.files[0].desc) {
                            loadState.data = IMatch.safeTagValue(response.files[0].desc);
                            $('#note').val(loadState.data);
                        }
                    }
                    else {
                        $('#note').val('');
                        loadState.data = '';
                    }
                },function(error) {
                    $('#note').val('');
                    loadState.data = '';
                    alert(IMatchTranslate.getStr('T_LOAD_ERROR'));
                });
            }
            else {
                // Load the Attribute
                IMWS.get('v1/attributes', {
                    id: loadState.focusedFile,
                    set: NOTES_ATTR_SET,
                    attributes: NOTES_ATTR  
                }).then(function(response) {
                    // Do we have a file?
                    if (response.result.length) {
                        // Does it have Attribute data?
                        if (response.result[0].data.length) {
                            // We use the first returned record
                            var data = response.result[0].data[0];
                            loadState.instanceId = data.instanceId;

                            // Only update if the data differs
                            if (loadState.data != data[NOTES_ATTR]) {
                                loadState.data = data[NOTES_ATTR];
                                $('#note').val(loadState.data);
                            }
                        }
                        else {
                            // No data yet
                            $('#note').val('');
                            loadState.data = '';
                        }
                    }
                    else {
                        // No data yet
                        $('#note').val('');
                        loadState.data = '';
                    }
                },function(error) {
                    $('#note').val('');
                    loadState.data = '';
                    alert(IMatchTranslate.getStr('T_LOAD_ERROR'));
                });
            }
        }
        else {
            // No file focused. We don't show anything
            $('#note').val('');
            loadState.data = '';
            $('#focused-file').text('');
            $('#sel-count').text('');
        }
    });  // idlist
}

